# Kornpob Bhirombhakdi
# kbhirombhakdi@stsci.edu

import numpy as np

class ConfReader:
    '''
    ConfReader (using Python 3) reads aXe configuration files.
    obj = ConfReader(conffile) to assign the configuration file
    obj.conffile to see what file
    obj.readall() simply reads the whole file
    obj.getbeam(beam) extracts the beam info
    obj.beam to see the beam info
    obj.make_coef2d() re-constructs DYDX and DLDP
    obj.coef2d to see the coef re-construction
    NOTE: use np.astype(type) to convert after extraction
    NOTE: can feed coef2d into Polynomial2D package (pip install polynomial2d) for further analysis
        (i.e., given X1,X2 arrays and obj.coef2d, Polynomial2D().fit() can decompress from 2D coef to 1D coef)
    '''
    def __init__(self,conffile=None):
        self.conffile = conffile
        self.beam = None
        self.coef2d = None
    ##########
    ##########
    ##########
    def getbeam(self,beam='A'):
        #####
        # prepare keywords to read
        KEY = ['BEAM'+beam,'MMAG_EXTRACT_'+beam,'MMAG_MARK_'+beam,
               'DYDX_ORDER_'+beam,
               'XOFF_'+beam,'YOFF_'+beam,
               'DISP_ORDER_'+beam
              ]
        #####
        # open file and record value for each keyword
        f = open(self.conffile,'r')
        tmpp = {}
        tmpp['BEAM'] = beam
        for i,ii in enumerate(f.readlines()):
            tmp = ii.split(' ')[0]
            if tmp in KEY:
                tmppp = np.array(ii.split()[1:])
                tmpp[tmp] = tmppp
        f.close()
        #####
        # get DYDX_{beam}_{order} and DLDP_{beam}_{order}
        tmpkey = {'DYDX_ORDER_'+beam:'DYDX_{0}_'.format(beam),
                  'DISP_ORDER_'+beam:'DLDP_{0}_'.format(beam)
                 }
        for i in tmpkey:
            order = int(tmpp[i][0])
            for j in np.arange(order+1):
                string = tmpkey[i]+str(j)
                f = open(self.conffile,'r')
                for k,kk in enumerate(f.readlines()):
                    if string in kk.split(' '):
                        tmpp[string] = np.array(kk.split()[1:])
        #####
        # keep in self
        self.beam = tmpp
    ##########
    ##########
    ##########
    def readall(self):
        f = open(self.conffile,'r')
        for i,ii in enumerate(f.readlines()):
            print(i,ii.split())
    ##########
    ##########
    ##########
    def make_coef2d(self):
        beam = self.beam['BEAM']
        KEY = {'DYDX_ORDER_'+beam:'DYDX_{0}_'.format(beam),
               'DISP_ORDER_'+beam:'DLDP_{0}_'.format(beam)
              }
        tmpp = {}
        for i in KEY:
            order = int(self.beam[i][0])
            for j in np.arange(order+1):
                string = KEY[i]+str(j)
                tmp = self.beam[string]
                tmpp[string] = self._make_coef(tmp)
        self.coef2d = tmpp
    def _make_coef(self,coef):
        tmp = np.array(coef).astype(float)
        px1,px2,order = 0,0,0 # initialize
        out = {}
        out['NORDER'] = None
        out['COEF'] = {}
        for i in tmp:
            out['COEF'][(px1,px2)] = i
            px1-=1
            px2+=1
            if px1 < 0:
                order += 1
                px1 = order
                px2 = 0
        out['NORDER'] = order - 1
        return out
    



# Kornpob Bhirombhakdi
# kbhirombhakdi@stsci.edu

from astropy.io import fits
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import copy
from .confreader import ConfReader
from .computesip import ComputeSIP
from .computetracenwavelength import ComputeTraceNWavelength

class Wfc3Uvis1G280:
    def __init__(self,objname='None',dfile=None,gfile=None,extnum=4,
                 conffile=None,flatfile=None,sensfile=None,beam=None,
                 xyd=None
                ):
        self.data = {'objname':objname,
                     'dfile':dfile,
                     'gfile':gfile,
                     'extnum':extnum,
                     'conffile':conffile,
                     'flatfile':flatfile,
                     'sensfile':sensfile,
                     'beam':beam,
                     'xyd':xyd,
                     'ROOTNAME':None,
                     'CCDCHIP':None,
                     'NAXIS1':None,
                     'NAXIS2':None,
                     'xgbound':(100.,200.)
                    }
        self.conf = None
        self.trace = {'XYREF':(None,None),
                      'XG':None,
                      'YG':None,
                      'WW':None
                     }
        self._get_root()
        self._get_ccdchip()
        self._get_nax()
        self._get_conf()
    def _get_root(self):
        try:
            tmp = fits.open(self.data['gfile'])[0].header['ROOTNAME']
            self.data['ROOTNAME'] = tmp
        except:
            print('Cannot find ROOTNAME for {0}.\n'.format(self.data['gfile']))
    def _get_ccdchip(self):
        try:
            EXTNUM = self.data['extnum']
            tmp = fits.open(self.data['gfile'])[EXTNUM].header['CCDCHIP']
            self.data['CCDCHIP'] = tmp
        except:
            print('Cannot find CCDCHIP for {0}.\n'.format(self.data['gfile']))
    def _get_nax(self):
        try:
            tmp = fits.open(self.data['gfile'])[self.data['extnum']].header
            self.data['NAXIS1'] = tmp['NAXIS1']
            self.data['NAXIS2'] = tmp['NAXIS2']
        except:
            print('Cannot find NAXIS1,NAXIS2 for {0}.\n'.format(self.data['gfile']))
    def _get_conf(self): 
        try:
            conffile = self.data['conffile']
            conf = ConfReader(conffile)
            conf.getbeam(beam=self.data['beam'])
            conf.make_coef2d()
            self.conf = copy.deepcopy(conf)
        except:
            print('Cannot get conf for {0}.\n'.format(self.data['gfile']))
    ##########
    ##########
    ##########
    def compute(self):
        xyd = self.data['xyd']
        xydiff = self._compute_xydiff()
        xyoff = self._get_xyoff()
        xyref = np.array(xyd) + np.array(xydiff) + np.array(xyoff)
        obj = ComputeSIP(self.conf.coef2d,xyref[0],xyref[1])
        obj.compress()
        newobj = ComputeTraceNWavelength(obj.coef1d,obj.x1,obj.x2,self.data['NAXIS1'],self.data['NAXIS2'])
        newobj.compute()
        tmpx,tmpy,tmpw = newobj.trace['XG'],newobj.trace['YG'],newobj.wavelength['WW']
        xref = xyref[0]
        xgbound = self.data['xgbound']
        m = np.where((tmpx >= xref-xgbound[0])&(tmpx <= xref+xgbound[1]))
        tmpx,tmpy,tmpw = tmpx[m],tmpy[m],tmpw[m]
        self.trace['XYREF'] = xyref
        self.trace['XG'] = tmpx
        self.trace['YG'] = tmpy
        self.trace['WW'] = tmpw
        self.trace['COEF1D'] = newobj.data['COEF1D']
    def _compute_xydiff(self):
        EXTNUM = self.data['extnum']
        tmpd = fits.open(self.data['dfile'])
        tmpd_p1,tmpd_p2,tmpd_s = tmpd[0].header['POSTARG1'],tmpd[0].header['POSTARG2'],tmpd[EXTNUM].header['IDCSCALE']
        tmpg = fits.open(self.data['gfile'])
        tmpg_p1,tmpg_p2,tmpg_s = tmpg[0].header['POSTARG1'],tmpg[0].header['POSTARG2'],tmpg[EXTNUM].header['IDCSCALE']
        dx = tmpg_p1/tmpg_s - tmpd_p1/tmpd_s
        dy = tmpg_p2/tmpg_s - tmpd_p2/tmpd_s
        return (dx,dy)
    def _get_xyoff(self):
        beam = self.data['beam']
        confbeam = self.conf.beam
        return (confbeam['XOFF_'+beam].astype(float)[0],confbeam['YOFF_'+beam].astype(float)[0])
    ##########
    ##########
    ##########
    def save(self,savename='trace.csv'):
        ##### xyref
        xyref = self.trace['XYREF']
        xref = np.full_like(self.trace['XG'],None,dtype=float)
        yref = np.full_like(self.trace['XG'],None,dtype=float)
        xref[0],yref[0] = xyref[0],xyref[1]
        ##### dydx, dldp
        dydx = np.full_like(self.trace['XG'],None,dtype=float)
        dldp = np.full_like(self.trace['XG'],None,dtype=float)
        for i in self.trace['COEF1D']:
            if i.split('_')[0] == 'DYDX':
                j = int(i.split('_')[-1])
                dydx[j] = self.trace['COEF1D'][i]
            elif i.split('_')[0] == 'DLDP':
                j = int(i.split('_')[-1])
                dldp[j] = self.trace['COEF1D'][i]
            else:
                pass
        tmp = {'XREF':xref,'YREF':yref,'XG':self.trace['XG'],'YG':self.trace['YG'],'WW':self.trace['WW'],
               'DYDX':dydx,'DLDP':dldp
              }
        pd.DataFrame(tmp).to_csv(savename,index=False)
        print('Save {0}'.format(savename))
    ##########
    ##########
    ##########
    def show(self,save=False,savename='plot.eps',saveformat='eps',
             dminmax=(5.,99.),gminmax=(5.,80.),alpha=0.6,lw=4,tickperx=50,
             dxy=(50,50),rotation=30,
             figsize=(10,10),fontsize=12
            ):
        EXTNUM = self.data['extnum']
        pixx,pixy = self.data['xyd'][0],self.data['xyd'][1]
        OBJNAME = self.data['objname']
        root = self.data['ROOTNAME']
        xg,yg,ww = self.trace['XG'],self.trace['YG'],self.trace['WW']
        xyref = self.trace['XYREF']
        dx,dy = dxy[0],dxy[1]        
        plt.figure(figsize=figsize)
        
        ax1 = plt.subplot(2,2,1)
        tmp = fits.open(self.data['dfile'])
        tmpheader = tmp[0].header
        tmpp = tmp[EXTNUM]
        tmppheader = tmpp.header
        tmppdata = tmpp.data
        vmin,vmax = np.percentile(tmppdata,dminmax[0]),np.percentile(tmppdata,dminmax[1])
        ax1.imshow(tmppdata,origin='lower',cmap='viridis',vmin=vmin,vmax=vmax)
        ax1.scatter(pixx,pixy,s=30,facecolor='None',edgecolor='red')
        fname = self.data['dfile'].split('/')[-1]
        string = '{0} {1} {2} SUBARRAY={3}\n'.format(fname,tmpheader['DATE-OBS'],tmpheader['FILTER'],tmpheader['SUBARRAY'])
        string += 'EXPSTART={0:.3f} EXPTIME={1:.3f}\n'.format(tmpheader['EXPSTART'],tmpheader['EXPTIME'])
        string += 'EXTNUM={0} CCDCHIP={1} BUNIT={2}'.format(EXTNUM,tmppheader['CCDCHIP'],tmppheader['BUNIT'])
        ax1.set_title(string,fontsize=fontsize)

        ax2 = plt.subplot(2,2,2)
        ax2.imshow(tmppdata,origin='lower',cmap='viridis',vmin=vmin,vmax=vmax)
        ax2.scatter(pixx,pixy,s=30,facecolor='None',edgecolor='red')
        ax2.set_xlim(pixx-dx,pixx+dx)
        ax2.set_ylim(pixy-dy,pixy+dy)
        string = '{0}\n'.format(OBJNAME)
        string += 'xy={0:.3f},{1:.3f}'.format(pixx,pixy)
        ax2.set_title(string,fontsize=fontsize)
        
        ax3 = plt.subplot(2,2,3)
        tmp = fits.open(self.data['gfile'])
        tmpheader = tmp[0].header
        tmpp = tmp[EXTNUM]
        tmppheader = tmpp.header
        tmppdata = tmpp.data
        vmin,vmax = np.percentile(tmppdata,gminmax[0]),np.percentile(tmppdata,gminmax[1])
        ax3.imshow(tmppdata,origin='lower',cmap='viridis',vmin=vmin,vmax=vmax)
        ax3.plot(xg,yg,'r-',alpha=alpha,lw=lw)
        fname = self.data['gfile'].split('/')[-1]
        string = '{0} {1} {2} SUBARRAY={3}\n'.format(fname,tmpheader['DATE-OBS'],tmpheader['FILTER'],tmpheader['SUBARRAY'])
        string += 'EXPSTART={0:.3f} EXPTIME={1:.3f}\n'.format(tmpheader['EXPSTART'],tmpheader['EXPTIME'])
        string += 'EXTNUM={0} CCDCHIP={1} BUNIT={2}'.format(EXTNUM,tmppheader['CCDCHIP'],tmppheader['BUNIT'])
        ax3.set_title(string,fontsize=fontsize)
        
        ax4 = plt.subplot(2,2,4)
        ax4.imshow(tmppdata,origin='lower',cmap='viridis',vmin=vmin,vmax=vmax)
        ax4.plot(xg,yg,'r:',alpha=alpha,lw=lw)
        for i,ii in enumerate(xg):
            if (i in {0,len(xg)-1}) or (np.mod(i,tickperx)==0):
                label = '{0}A'.format(int(ww[i]))
                ax4.plot(xg[i],yg[i],'ro')
                ax4.annotate(label,(xg[i],yg[i]),
                             textcoords='offset points',
                             xytext=(0,10),
                             ha='center',
                             fontsize=fontsize,
                             rotation=rotation
                            )
        ax4.set_xlim(xg.min()-dx,xg.max()+dx)
        ax4.set_ylim(yg.min()-dy,yg.max()+dy)  
        string = 'xyref={0:.3f},{1:.3f}'.format(xyref[0],xyref[1])
        ax4.set_title(string,fontsize=fontsize)
        
        plt.tight_layout()
        if save:
            plt.savefig(savename,format=saveformat,bbox_inches='tight')
            print('Save {0}\n'.format(savename))
        



# Kornpob Bhirombhakdi
# kbhirombhakdi@stsci.edu

import numpy as np
from scipy.integrate import quad

class ComputeTraceNWavelength:
    '''
    ComputeTraceNWavelength (using Python 3) computes trace and wavelength in HST/aXe definition.
    This class can be easily used following ComputeSIP.
    Given obj = ComputeSIP(conf.coef2d,xref,yref), simply call
        newobj = ComputeTraceNWavelength(obj.coef1d,obj.x1,obj.x2,nax1,nax2)
        where nax1,nax2 = dimension of a grism image.
    newobj.data to see input parameters.
    newobj.compute() to compute trace and wavelength.
    newobj.trace to see trace
    newobj.wavelength to see wavelength (in A)
    '''
    def __init__(self,coef1d=None,x1ref=None,x2ref=None,
                 nax1=None,nax2=None
                ):
        self.data = {'COEF1D':coef1d,
                     'X1REF':x1ref,
                     'X2REF':x2ref,
                     'NX1':nax1,
                     'NX2':nax2,
                    }
        self.trace = None
        self.wavelength = None
    ##########
    ##########
    ##########
    def compute(self):
        # trace
        xg = np.arange(self.data['NX1'])
        delx = xg - self.data['X1REF']
        dydx = self._get_coef('DYDX')
        tmp = 0.
        for i,ii in enumerate(dydx):
            tmp += ii * np.power(delx,i)
        self.trace = {'XG':xg,
                      'YG':tmp.copy() + self.data['X2REF']
                     }
        # wavelength
        dydx = self._get_coef('DYDX')
        dldp = self._get_coef('DLDP')
        varclength = np.vectorize(self._arclength)
        arc,earc = np.array(varclength(delx,*dydx))
        tmp = 0.
        for i,ii in enumerate(dldp):
            tmp += ii * np.power(arc,i)
        self.wavelength = {'XG':xg,
                           'WW':tmp.copy()
                          }
    def _get_coef(self,key):
        coef1d = self.data['COEF1D']
        tmp = {}
        for i in coef1d:
            if i.split('_')[0] == key:
                tmp[int(i.split('_')[2])] = coef1d[i]
        n = len(tmp)
        dydx = []
        for i in range(n):
            dydx.append(tmp[i])
        return dydx
    def _arclength_integrand(self,Fa,*coef):
        # compute np.sqrt(1 + np.power(dydx,2))
        # dy = sum(i=0,i=n,DYDX_A_i*np.power(dx,i))
        # dydx = (dy/dx) = sum(i=0,i=n,i*DYDX_A_i*np.power(dx,i-1))
        # Fa = dx
        # coef = [DYDX_A_0,DYDX_A_1,...]
        s = 0
        for i,ii in enumerate(coef):
            if i==0:
                continue
            s += i * ii * np.power(Fa,i-1)
        return np.sqrt(1. + np.power(s,2))
    def _arclength(self,Fa,*coef):
        # compute integrate(from 0,to Fa,integrand np.sqrt(1 + np.power(dydx,2)))
        integral,err = quad(self._arclength_integrand, 0., Fa, args=coef)
        return integral,err 



# Kornpob Bhirombhakdi
# kbhirombhakdi@stsci.edu

from polynomial2d.polynomial2d import Polynomial2D

class ComputeSIP:
    '''
    ComputeSIP (using Python 3) chains from ConfReader to compute SIP using Polynomial2D package (pip install polynomial2d).
    conf = ConfReader() with conf.coef2d corrently constructed.
    obj = ComputeSIP(conf.coef2d,x1=xref,x2=yref) to instantiate (NOTE: x1 is the leading term in SIP).
    obj.coef2d contains the same conf.coef2d
    obj.x1,obj.x2 contains xref,yref
    obj.compress() to compress to 1D coef
    self.coef1d to see the 1D coef
    '''
    def __init__(self,coef2d=None,x1=None,x2=None):
        self.coef2d = coef2d
        self.x1 = x1
        self.x2 = x2
    def compress(self):
        out = {}
        for i in self.coef2d:
            tmp = self.coef2d[i]   
            obj = Polynomial2D()
            obj.model['NORDER'] = tmp['NORDER']
            obj.model['COEF'] = tmp['COEF']
            obj.data['X1'] = self.x1
            obj.data['X2'] = self.x2
            obj.compute()
            out[i] = obj.model['YFIT']
        self.coef1d = out           



# Kornpob Bhirombhakdi
# kbhirombhakdi@stsci.edu

from astropy.io import fits
import copy
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

class FlatField:
    '''
    FlatField (using Python 3) computes a flatfield image in aXe definition.
    This can easily be implemented following the use of ComputeTraceNWavelength.
    Given obj = ComputeTraceNWavelength(...), and obj.compute() is called,
        a user can start with flat = FlatField(flatfile,obj.wavelength['WW'])
    flat.flatfile to see the path to flatfile.
    flat.data to see inputs for the computation.
    flat.compute() to compute the flatfield image.
    flat.flatfield to see the flatfield image array.
    flat.show(...) to imshow flatfield.
    flat.save(...) to save as *_flat.fits.
    '''
    def __init__(self,flatfile,tracefile):
        self.flatfile = flatfile
        self.tracefile = tracefile
        self.data = {'WMIN':None,
                     'WMAX':None,
                     'FLAT':None,
                     'NAX1':None,
                     'NAX2':None,
                     'WW':None,
                     'ROOTNAME':None
                    }
        self.flatfield = None
        self._get()
    def _get(self):
        # basic info
        tmp = fits.open(self.flatfile)
        wmin,wmax = tmp[0].header['WMIN'],tmp[0].header['WMAX']
        nax1,nax2 = tmp[0].header['NAXIS1'],tmp[0].header['NAXIS2']
        self.data['WMIN'] = wmin
        self.data['WMAX'] = wmax
        self.data['NAX1'] = nax1
        self.data['NAX2'] = nax2
        self.data['ROOTNAME'] = self.tracefile.split('/')[-1].split('_')[0]
        # flat
        n = len(tmp)
        tmpp = {}
        for i in range(n):
            tmpp[i] = tmp[i].data.copy()
        self.data['FLAT'] = copy.deepcopy(tmpp)
        # wavelength
        tmp = pd.read_csv(self.tracefile)
        dydx = tmp['DYDX']
        m = np.isfinite(dydx)
        dydx = dydx[m].values
        xg = np.arange(self.data['NAX1'])
        xref = tmp['XREF'][0]
        xh = xg-xref
        ww = np.full_like(xh,0.,dtype=float)
        for i,ii in enumerate(dydx):
            ww += ii * np.power(xh,i)
        self.data['WW'] = ww
    ##########
    ##########
    ##########
    def compute(self):
        # dimension test
        if self.data['WW'].shape[0] != self.data['NAX1']:
            raise ValueError('Dimension of WW != NAX1')
        # set parametric wavelength [0,1]
        wmin,wmax,ww = self.data['WMIN'],self.data['WMAX'],self.data['WW']
        paramww = (ww-wmin)/(wmax-wmin)
        paramww[paramww<0.] = 0.
        paramww[paramww>1.] = 1.
        paramww,_ = np.meshgrid(paramww,np.arange(self.data['NAX2']))
        # compute
        tmp = self.data['FLAT']
        tmpp = np.full_like(paramww,0.,dtype=float)
        for i in tmp:
            tmpp += tmp[i] * np.power(paramww,i)
        self.flatfield = tmpp.copy()
    ##########
    ##########
    ##########
    def save(self,savename_prefix='image',overwrite=False):
        hdu = fits.PrimaryHDU()
        hdul = fits.HDUList([hdu])
        hdul[0].header['NAXIS'] = 2
        hdul[0].header['EXTEND'] = False
        hdul[0].header['DESC'] = 'Flatfield for hstgrism'
        hdul[0].header['NAX1'] = self.data['NAX1']
        hdul[0].header['NAX2'] = self.data['NAX2']
        hdul[0].header['FLAT'] = self.flatfile
        hdul[0].header['TRACE'] = self.tracefile
        hdul[0].header['ROOTNAME'] = self.data['ROOTNAME']
        hdul[0].data = self.flatfield.copy()
        string = '{0}_flat.fits'.format(savename_prefix)
        hdul.writeto(string,overwrite=overwrite)
        print('Save {0}'.format(string))
    ##########
    ##########
    ##########
    def show(self,
             save=False,savename_prefix='image',saveformat='eps',
             params={'figsize':(5,5),
                     'minmax':(5.,99.),
                     'cmap':'viridis',
                     'fontsize':12
                    }
            ):
        figsize = params['figsize']
        minmax = params['minmax']
        cmap = params['cmap']
        fontsize = params['fontsize']
        plt.figure(figsize=figsize)
        tmpp = self.flatfield
        vmin,vmax = np.percentile(tmpp,minmax[0]),np.percentile(tmpp,minmax[1])
        plt.imshow(self.flatfield,origin='lower',cmap=cmap,vmin=vmin,vmax=vmax)
        plt.title(self.data['ROOTNAME'],fontsize=fontsize)
        plt.tight_layout()
        if save:
            string = '{0}_flat.{1}'.format(savename_prefix,saveformat)
            plt.savefig(string,format=saveformat,bbox_inches='tight')
            print('Save {0}'.format(string))
    
    


# Kornpob Bhirombhakdi
# kbhirombhakdi@stsci.edu

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from astropy.io import fits
from polynomial2d.polynomial2d import Polynomial2D
import os
import json

class Background:
    def __init__(self,objname='None',gfile='None',tfile='None',extnum=1,
                 padxleft=5,xleft=100,xright=200,padxright=5,
                 padyup=5,halfdy=5,padylow=5,
                 adjustx=0,adjusty=0
                ):
        self.data = {'objname':objname,
                     'gfile':gfile,
                     'tfile':tfile,
                     'extnum':extnum,
                     'ROOTNAME':'None',
                    }
        self.data['ROOTNAME'] = self._rootname()
        self.trace = self._trace()
        self.bbox = {'padxleft':padxleft,'xleft':xleft,'xright':xright,'padxright':padxright,
                     'padyup':padyup,'halfdy':halfdy,'padylow':padylow,
                     'adjustx':adjustx,'adjusty':adjusty,
                     'bbox':None
                     }
        self.bbox['bbox'] = self._bbox()
        self.bkg = self._bkg()
    def _bkg(self):
        tmpdata = fits.open(self.data['gfile'])[self.data['extnum']].data
        bb0x,bb1x = self.bbox['bbox']['BBX']
        bb0y,bb1y = self.bbox['bbox']['BBY']
        xg,yg = self.bbox['bbox']['XG'],self.bbox['bbox']['YG']
        tmpp = tmpdata[bb0y:bb1y,bb0x:bb1x]
        ny,nx = tmpp.shape
        x1 = np.arange(nx)
        x2 = np.arange(ny)
        x1v,x2v = np.meshgrid(x1,x2)
        obj = Polynomial2D(x1v,x2v,tmpp)
        obj.data['MASK'] = self._mask(obj)
        return obj
    def _mask(self,obj):
        tmpmask = np.full_like(obj.data['Y'],False,dtype=bool)
        halfdy = self.bbox['halfdy']
        xgn = (self.bbox['bbox']['XG'] - self.bbox['bbox']['BBX'][0]).astype(int)
        ygn = (self.bbox['bbox']['YG'] - self.bbox['bbox']['BBY'][0]).astype(int)
        for i,ii in enumerate(xgn):
            tmpx,tmpy = int(xgn[i]),int(ygn[i])
            tmpmask[tmpy-halfdy:tmpy+halfdy,tmpx] = True
        return tmpmask
    def _rootname(self):
        tmp = fits.open(self.data['gfile'])[0].header['ROOTNAME']
        return tmp
    def _trace(self):
        tmp = pd.read_csv(self.data['tfile'])
        mdydx = np.isfinite(tmp['DYDX'].values)
        mdldp = np.isfinite(tmp['DLDP'].values)
        tmpp = {'XREF':tmp['XREF'].values[0],
                'YREF':tmp['YREF'].values[0],
                'XG':tmp['XG'].values,
                'YG':tmp['YG'].values,
                'WW':tmp['WW'].values,
                'DYDX':tmp['DYDX'].values[mdydx],
                'DLDP':tmp['DLDP'].values[mdldp]
               }
        return tmpp
    def _bbox(self):
        xref,yref = self.trace['XREF'],self.trace['YREF']
        maxy,miny = self.trace['YG'].max(),self.trace['YG'].min()
        bb0y,bb1y = int(miny-self.bbox['halfdy']-self.bbox['padylow']),int(maxy+self.bbox['halfdy']+self.bbox['padyup'])
        maxx,minx = self.trace['XG'].max(),self.trace['XG'].min()
        bb0x,bb1x = int(minx-self.bbox['padxleft']),int(maxx+self.bbox['padxright'])
        xg = np.arange(bb0x,bb1x)
        xh = xg - int(xref)
        dydx = self.trace['DYDX']
        dldp = self.trace['DLDP']
        yh = np.full_like(xh,0.,dtype=float)
        for i,ii in enumerate(dydx):
            yh += ii * np.power(xh,i)
        yg = yh + yref
        ww = np.full_like(xh,0.,dtype=float)
        for i,ii in enumerate(dldp):
            ww += ii * np.power(xh,i)
        return {'BBX':(bb0x,bb1x),'BBY':(bb0y,bb1y),'XG':xg,'YG':yg,'WW':ww}   
    ##########
    ##########
    ##########
    def saveinfo(self,savefolder='EXTRA'):
        os.system('mkdir {0}'.format(savefolder))
        # save bkg sub
        string = './{1}/{0}'.format(self.data['gfile'].split('/')[-1],savefolder)
        os.system('cp {0} ./{1}/'.format(self.data['gfile'],savefolder))
        tmpp = fits.open(string)
        tmppdata = tmpp[self.data['extnum']].data
        bb0x,bb1x = self.bbox['bbox']['BBX']
        bb0y,bb1y = self.bbox['bbox']['BBY']
        y,yfit = self.bkg.data['Y'],self.bkg.model['YFIT']
        ysub = y-yfit
        tmppdata[bb0y:bb1y,bb0x:bb1x] = ysub.copy()
        tmpp.writeto(string,overwrite=True)
        print('Save {0}'.format(string))
        # save bbox
        tmpp = {}
        for j in self.bbox:
            if j!='bbox':
                tmpp[j] = self.bbox[j]
            elif j=='bbox':
                tmpp['BBX'] = self.bbox[j]['BBX']
                tmpp['BBY'] = self.bbox[j]['BBY']
        tmppp = json.dumps(tmpp)
        string = './{1}/{0}_bbox.json'.format(self.data['ROOTNAME'],savefolder)
        f = open(string,'w')
        f.write(tmppp)
        f.close()
        print('Save {0}'.format(string))
    ##########
    ##########
    ##########
    def show_bkg2dsub(self,
                      save=False,savename_prefix='plot',saveformat='eps',
                      params={'figsize':(30,10),
                              'minmax_y':(10.,99.),
                              'minmax_ysub':(5.,99.),
                              'cmap':'rainbow',
                              'fontsize':12
                             }
                     ):
        figsize = params['figsize']
        minmax_y = params['minmax_y']
        minmax_ysub = params['minmax_ysub']
        cmap = params['cmap']
        fontsize = params['fontsize']
        y,yfit = self.bkg.data['Y'],self.bkg.model['YFIT']
        ysub = y-yfit
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(1,3,1)
        vmin,vmax = np.percentile(y,minmax_y[0]),np.percentile(y,minmax_y[1])
        ax.imshow(y,origin='lower',cmap=cmap,vmin=vmin,vmax=vmax)
        ax.set_title(self.data['ROOTNAME'],fontsize=fontsize)
        ax = fig.add_subplot(1,3,2)
        ax.imshow(yfit,origin='lower',cmap=cmap)
        ax.set_title('fit',fontsize=fontsize)
        ax = fig.add_subplot(1,3,3)
        vmin,vmax = np.percentile(ysub,minmax_ysub[0]),np.percentile(ysub,minmax_ysub[1])
        ax.imshow(ysub,origin='lower',cmap=cmap,vmin=vmin,vmax=vmax)
        ax.set_title('sub',fontsize=fontsize)
        fig.tight_layout()               
        if save:
            string = '{0}_bkg2dsub.{1}'.format(savename_prefix,saveformat)
            plt.savefig(string,format=saveformat,bbox_inches='tight')
            print('Save {0}'.format(string))
    def show_bkg3d(self,
                   save=False,savename_prefix='plot',saveformat='eps',
                   params={'figsize':(20,10),
                           'cmap':'rainbow',
                           'view_init_y':(90.,-90.),
                           'view_init_mask':(90.,-90.),
                           'fontsize':12
                          }
                  ):
        figsize = params['figsize']
        cmap = params['cmap']
        view_init_y = params['view_init_y']
        view_init_mask = params['view_init_mask']
        fontsize = params['fontsize']
        x1,x2,y,mask = self.bkg.data['X1'],self.bkg.data['X2'],self.bkg.data['Y'],self.bkg.data['MASK']
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(1,2,1,projection='3d')
        ax.plot_surface(x1,x2,y,cmap=cmap)
        ax.view_init(*view_init_y)
        ax.set_title(self.data['ROOTNAME'],fontsize=fontsize)
        ax = fig.add_subplot(1,2,2,projection='3d')
        ax.plot_surface(x1,x2,mask,cmap=cmap)
        ax.view_init(*view_init_mask)
        ax.set_title('MASK',fontsize=fontsize)
        fig.tight_layout()               
        if save:
            string = '{0}_bkg3d.{1}'.format(savename_prefix,saveformat)
            plt.savefig(string,format=saveformat,bbox_inches='tight')
            print('Save {0}'.format(string))
    def show_bkg2d(self,
                   save=False,savename_prefix='plot',saveformat='eps',
                   params={'figsize':(10,10),
                           'cmap_mask':'Greys',
                           'cmap_y':'viridis',
                           'alpha':0.4,
                           'color':'red',
                           'ls':'-',
                           'lw':2,
                           'fontsize':12
                          }
                  ):
        figsize = params['figsize']
        cmap_mask = params['cmap_mask']
        cmap_y = params['cmap_y']
        alpha = params['alpha']
        color = params['color']
        ls = params['ls']
        lw = params['lw']
        fontsize = params['fontsize']
        plt.figure(figsize=figsize)
        plt.imshow(self.bkg.data['MASK'],origin='lower',cmap=cmap_mask)
        plt.imshow(self.bkg.data['Y'],origin='lower',cmap=cmap_y,alpha=alpha)
        tmpx = self.bbox['bbox']['XG'] - self.bbox['bbox']['BBX'][0]
        tmpy = self.bbox['bbox']['YG'] - self.bbox['bbox']['BBY'][0]
        plt.plot(tmpx,tmpy,color=color,ls=ls,lw=lw)
        root = self.data['ROOTNAME']
        plt.title(root,fontsize=fontsize)
        plt.tight_layout()
        if save:
            string = '{0}_bkg2d.{1}'.format(savename_prefix,saveformat)
            plt.savefig(string,format=saveformat,bbox_inches='tight')
            print('Save {0}'.format(string))
    def show_bbox2d(self,
                    save=False,savename_prefix='plot',saveformat='eps',
                    params={'figsize':(10,10),
                            'minmax':(10.,80.),
                            'tickperx':10,
                            'fontsize':12,
                            'rotation':30.,
                            'color':'r',
                            'cmap':'viridis',
                            'lw':2,
                            'ls':':',
                            'marker':'o',
                            'alpha':0.6,
                            'border_x':50,
                            'border_y':50
                           }
                   ):
        tmpdata = fits.open(self.data['gfile'])[self.data['extnum']].data
        minmax = params['minmax']
        cmap = params['cmap']
        figsize = params['figsize']
        color = params['color']
        ls = params['ls']
        lw = params['lw']
        alpha = params['alpha']
        border_x = params['border_x']
        border_y = params['border_y']
        fontsize = params['fontsize']
        tickperx = params['tickperx']
        marker = params['marker']
        rotation = params['rotation']
        vmin,vmax = np.percentile(tmpdata,minmax[0]),np.percentile(tmpdata,minmax[1])
        xg,yg,ww = self.bbox['bbox']['XG'],self.bbox['bbox']['YG'],self.bbox['bbox']['WW']
        bb0x,bb1x = self.bbox['bbox']['BBX']
        bb0y,bb1y = self.bbox['bbox']['BBY']
        plt.figure(figsize=figsize)
        plt.imshow(tmpdata,origin='lower',cmap=cmap,vmin=vmin,vmax=vmax)
        plt.plot(xg,yg,color=color,ls=ls,lw=lw,alpha=alpha)
        for i,ii in enumerate(xg):
            if (i in {0,len(xg)-1}) or (np.mod(i,tickperx)==0):
                label = '{0}A'.format(int(ww[i]))
                plt.plot(xg[i],yg[i],color=color,marker=marker)
                plt.annotate(label,(xg[i],yg[i]),
                             textcoords='offset points',
                             xytext=(0,10),
                             ha='center',
                             fontsize=fontsize,
                             rotation=rotation,
                             color=color
                            )
        plt.plot([bb0x,bb1x,bb1x,bb0x,bb0x],[bb0y,bb0y,bb1y,bb1y,bb0y],color=color,ls='-')
        plt.xlim(bb0x-border_x,bb1x+border_x)
        plt.ylim(bb0y-border_y,bb1y+border_y)
        plt.title(self.data['ROOTNAME'],fontsize=fontsize) 
        plt.tight_layout()
        if save:
            string = '{0}_bbox2d.{1}'.format(savename_prefix,saveformat)
            plt.savefig(string,format=saveformat,bbox_inches='tight')
            print('Save {0}'.format(string))
            
            